/*
 * QEMU Block driver for  NBD
 *
 * Copyright (C) 2016 Red Hat, Inc.
 * Copyright (C) 2008 Bull S.A.S.
 *     Author: Laurent Vivier <Laurent.Vivier@bull.net>
 *
 * Some parts:
 *    Copyright (C) 2007 Anthony Liguori <anthony@codemonkey.ws>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "qemu/osdep.h"
#include "nbd-client.h"

#define HANDLE_TO_INDEX(bs, handle) ((handle) ^ ((uint64_t)(intptr_t)bs))
#define INDEX_TO_HANDLE(bs, index)  ((index)  ^ ((uint64_t)(intptr_t)bs))

static void nbd_recv_coroutines_enter_all(NBDClientSession *s)
{
    int i;

    for (i = 0; i < MAX_NBD_REQUESTS; i++) {
        NBDClientRequest *req = &s->requests[i];

        if (req->coroutine && req->receiving) {
            aio_co_wake(req->coroutine);
        }
    }
}

static void nbd_teardown_connection(BlockDriverState *bs)
{
    NBDClientSession *client = nbd_get_client_session(bs);

    if (!client->ioc) { /* Already closed */
        return;
    }

    /* finish any pending coroutines */
    qio_channel_shutdown(client->ioc,
                         QIO_CHANNEL_SHUTDOWN_BOTH,
                         NULL);
    BDRV_POLL_WHILE(bs, client->read_reply_co);

    nbd_client_detach_aio_context(bs);
    object_unref(OBJECT(client->sioc));
    client->sioc = NULL;
    object_unref(OBJECT(client->ioc));
    client->ioc = NULL;
}

static coroutine_fn void nbd_read_reply_entry(void *opaque)
{
    NBDClientSession *s = opaque;
    uint64_t i;
    int ret = 0;

    while (!s->quit) {
        assert(s->reply.handle == 0);
        ret = nbd_receive_reply(s->ioc, &s->reply);
        if (ret <= 0) {
            break;
        }

        /* There's no need for a mutex on the receive side, because the
         * handler acts as a synchronization point and ensures that only
         * one coroutine is called until the reply finishes.
         */
        i = HANDLE_TO_INDEX(s, s->reply.handle);
        if (i >= MAX_NBD_REQUESTS ||
            !s->requests[i].coroutine ||
            !s->requests[i].receiving) {
            break;
        }

        /* We're woken up again by the request itself.  Note that there
         * is no race between yielding and reentering read_reply_co.  This
         * is because:
         *
         * - if the request runs on the same AioContext, it is only
         *   entered after we yield
         *
         * - if the request runs on a different AioContext, reentering
         *   read_reply_co happens through a bottom half, which can only
         *   run after we yield.
         */
        aio_co_wake(s->requests[i].coroutine);
        qemu_coroutine_yield();
    }

    s->quit = true;
    nbd_recv_coroutines_enter_all(s);
    s->read_reply_co = NULL;
}

static int nbd_co_send_request(BlockDriverState *bs,
                               NBDRequest *request,
                               QEMUIOVector *qiov)
{
    NBDClientSession *s = nbd_get_client_session(bs);
    int rc, ret, i;

    qemu_co_mutex_lock(&s->send_mutex);
    while (s->in_flight == MAX_NBD_REQUESTS) {
        qemu_co_queue_wait(&s->free_sema, &s->send_mutex);
    }
    s->in_flight++;

    for (i = 0; i < MAX_NBD_REQUESTS; i++) {
        if (s->requests[i].coroutine == NULL) {
            break;
        }
    }

    g_assert(qemu_in_coroutine());
    assert(i < MAX_NBD_REQUESTS);

    s->requests[i].coroutine = qemu_coroutine_self();
    s->requests[i].receiving = false;

    request->handle = INDEX_TO_HANDLE(s, i);

    if (s->quit) {
        qemu_co_mutex_unlock(&s->send_mutex);
        return -EIO;
    }
    if (!s->ioc) {
        qemu_co_mutex_unlock(&s->send_mutex);
        return -EPIPE;
    }

    if (qiov) {
        qio_channel_set_cork(s->ioc, true);
        rc = nbd_send_request(s->ioc, request);
        if (rc >= 0 && !s->quit) {
            ret = nbd_wr_syncv(s->ioc, qiov->iov, qiov->niov, request->len,
                               false);
            if (ret != request->len) {
                rc = -EIO;
            }
        }
        qio_channel_set_cork(s->ioc, false);
    } else {
        rc = nbd_send_request(s->ioc, request);
    }
    if (rc < 0) {
        s->quit = true;
    } else if (s->quit) {
        rc = -EIO;
    }
    qemu_co_mutex_unlock(&s->send_mutex);
    return rc;
}

static inline ssize_t read_sync(QIOChannel *ioc, void *buffer, size_t size)
{
    struct iovec iov = { .iov_base = buffer, .iov_len = size };
    /* Sockets are kept in blocking mode in the negotiation phase.  After
     * that, a non-readable socket simply means that another thread stole
     * our request/reply.  Synchronization is done with recv_coroutine, so
     * that this is coroutine-safe.
     */

    return nbd_wr_syncv(ioc, &iov, 1, size, true);
}

static void nbd_co_receive_reply(NBDClientSession *s,
    NBDRequest *request, NBDReply *reply,
    QEMUIOVector *qiov,
    NBDExtent **pextents, unsigned *nb_extents)
{
    int ind = HANDLE_TO_INDEX(s, request->handle);
    ssize_t ret;
    unsigned i;

    if (s->quit) {
        reply->error = EIO;
        return;
    }

    /* Wait until we're woken up by nbd_read_reply_entry.  */
    s->requests[ind].receiving = true;
    qemu_coroutine_yield();
    s->requests[ind].receiving = false;
    *reply = s->reply;
    if (reply->handle != request->handle || !s->ioc || s->quit) {
        reply->simple = true;
        reply->error = EIO;
    } else {
        if (qiov) {
            if ((reply->simple ? reply->error == 0 :
                         reply->type == NBD_REPLY_TYPE_OFFSET_DATA)) {
                ret = nbd_wr_syncv(s->ioc, qiov->iov, qiov->niov, request->len,
                                   true);
                if (ret != request->len || s->quit) {
                    reply->error = EIO;
                    s->quit = true;
                    return;
                }
            } else if (!reply->simple &&
                       reply->type == NBD_REPLY_TYPE_OFFSET_HOLE) {
                qemu_iovec_memset(qiov, 0, 0, request->len);
            }
        }

        if (reply->error == 0 && !reply->simple &&
            reply->type == NBD_REPLY_TYPE_BLOCK_STATUS &&
            pextents && nb_extents)
        {
            ret = read_sync(s->ioc, &reply->context_id,
                            sizeof(reply->context_id));
            if (ret != sizeof(reply->context_id) || s->quit) {
                reply->error = EIO;
                s->quit = true;
                return;
            }
            cpu_to_be32s(&reply->context_id);

            *nb_extents =
                (reply->length - sizeof(reply->context_id)) / sizeof(NBDExtent);
            *pextents = g_new(NBDExtent, *nb_extents);
            ret = read_sync(s->ioc, *pextents, *nb_extents * sizeof(NBDExtent));
            if (ret != *nb_extents * sizeof(NBDExtent) || s->quit) {
                g_free(*pextents);
                *pextents = NULL;
                reply->error = EIO;
                s->quit = true;
                return;
            }

            for (i = 0; i < *nb_extents; ++i) {
                cpu_to_be32s(&((*pextents)[i].length));
                cpu_to_be32s(&((*pextents)[i].flags));
            }
        }

        /* Tell the read handler to read another header.  */
        s->reply.handle = 0;
    }
}

static void nbd_coroutine_end(BlockDriverState *bs,
                              NBDRequest *request)
{
    NBDClientSession *s = nbd_get_client_session(bs);
    int i = HANDLE_TO_INDEX(s, request->handle);

    s->requests[i].coroutine = NULL;

    /* Kick the read_reply_co to get the next reply.  */
    if (s->read_reply_co) {
        aio_co_wake(s->read_reply_co);
    }

    qemu_co_mutex_lock(&s->send_mutex);
    s->in_flight--;
    qemu_co_queue_next(&s->free_sema);
    qemu_co_mutex_unlock(&s->send_mutex);
}

int nbd_client_co_preadv(BlockDriverState *bs, uint64_t offset,
                         uint64_t bytes, QEMUIOVector *qiov, int flags)
{
    NBDClientSession *client = nbd_get_client_session(bs);
    NBDRequest request = {
        .type = NBD_CMD_READ,
        .from = offset,
        .len = bytes,
        .flags = client->structured_reply ? NBD_CMD_FLAG_DF : 0,
    };
    NBDReply reply;
    ssize_t ret;

    assert(bytes <= NBD_MAX_BUFFER_SIZE);
    assert(!flags);

    ret = nbd_co_send_request(bs, &request, NULL);
    if (ret < 0) {
        goto out;
    }

    nbd_co_receive_reply(client, &request, &reply, qiov, NULL, NULL);
    if (reply.error != 0) {
        ret = -reply.error;
    }

    if (!reply.simple) {
        while (!(reply.flags & NBD_REPLY_FLAG_DONE)) {
            if (client->read_reply_co) {
                aio_co_wake(client->read_reply_co);
            } else {
                ret = -EIO;
                goto out;
            }
            nbd_co_receive_reply(client, &request, &reply, qiov,
                                 NULL, NULL);
            if (reply.error != 0) {
                ret = -reply.error;
            }
            if (reply.simple) {
                ret = -EIO;
                goto out;
            }
        }
    }

out:
    nbd_coroutine_end(bs, &request);

    return ret;
}

int nbd_client_co_pwritev(BlockDriverState *bs, uint64_t offset,
                          uint64_t bytes, QEMUIOVector *qiov, int flags)
{
    NBDClientSession *client = nbd_get_client_session(bs);
    NBDRequest request = {
        .type = NBD_CMD_WRITE,
        .from = offset,
        .len = bytes,
    };
    NBDReply reply;
    ssize_t ret;

    if (flags & BDRV_REQ_FUA) {
        assert(client->nbdflags & NBD_FLAG_SEND_FUA);
        request.flags |= NBD_CMD_FLAG_FUA;
    }

    assert(bytes <= NBD_MAX_BUFFER_SIZE);

    ret = nbd_co_send_request(bs, &request, qiov);
    if (ret < 0) {
        reply.error = -ret;
    } else {
        nbd_co_receive_reply(client, &request, &reply, NULL, NULL, NULL);
    }
    nbd_coroutine_end(bs, &request);
    return -reply.error;
}

int nbd_client_co_pwrite_zeroes(BlockDriverState *bs, int64_t offset,
                                int count, BdrvRequestFlags flags)
{
    ssize_t ret;
    NBDClientSession *client = nbd_get_client_session(bs);
    NBDRequest request = {
        .type = NBD_CMD_WRITE_ZEROES,
        .from = offset,
        .len = count,
    };
    NBDReply reply;

    if (!(client->nbdflags & NBD_FLAG_SEND_WRITE_ZEROES)) {
        return -ENOTSUP;
    }

    if (flags & BDRV_REQ_FUA) {
        assert(client->nbdflags & NBD_FLAG_SEND_FUA);
        request.flags |= NBD_CMD_FLAG_FUA;
    }
    if (!(flags & BDRV_REQ_MAY_UNMAP)) {
        request.flags |= NBD_CMD_FLAG_NO_HOLE;
    }

    ret = nbd_co_send_request(bs, &request, NULL);
    if (ret < 0) {
        reply.error = -ret;
    } else {
        nbd_co_receive_reply(client, &request, &reply, NULL, NULL, NULL);
    }
    nbd_coroutine_end(bs, &request);
    return -reply.error;
}

int nbd_client_co_flush(BlockDriverState *bs)
{
    NBDClientSession *client = nbd_get_client_session(bs);
    NBDRequest request = { .type = NBD_CMD_FLUSH };
    NBDReply reply;
    ssize_t ret;

    if (!(client->nbdflags & NBD_FLAG_SEND_FLUSH)) {
        return 0;
    }

    request.from = 0;
    request.len = 0;

    ret = nbd_co_send_request(bs, &request, NULL);
    if (ret < 0) {
        reply.error = -ret;
    } else {
        nbd_co_receive_reply(client, &request, &reply, NULL, NULL, NULL);
    }
    nbd_coroutine_end(bs, &request);
    return -reply.error;
}

int nbd_client_co_pdiscard(BlockDriverState *bs, int64_t offset, int count)
{
    NBDClientSession *client = nbd_get_client_session(bs);
    NBDRequest request = {
        .type = NBD_CMD_TRIM,
        .from = offset,
        .len = count,
    };
    NBDReply reply;
    ssize_t ret;

    if (!(client->nbdflags & NBD_FLAG_SEND_TRIM)) {
        return 0;
    }

    ret = nbd_co_send_request(bs, &request, NULL);
    if (ret < 0) {
        reply.error = -ret;
    } else {
        nbd_co_receive_reply(client, &request, &reply, NULL, NULL, NULL);
    }
    nbd_coroutine_end(bs, &request);
    return -reply.error;
}

static int nbd_client_co_cmd_block_status(BlockDriverState *bs, uint64_t offset,
                                          uint64_t bytes, NBDExtent **pextents,
                                          unsigned *nb_extents)
{
    int64_t ret;
    NBDReply reply;
    unsigned nb;
    NBDExtent *extents = NULL;
    NBDClientSession *client = nbd_get_client_session(bs);
    NBDRequest request = {
        .type = NBD_CMD_BLOCK_STATUS,
        .from = offset,
        .len = bytes,
        .flags = 0,
    };

    ret = nbd_co_send_request(bs, &request, NULL);
    if (ret < 0) {
        goto fail;
    }

    nbd_co_receive_reply(client, &request, &reply, NULL, &extents, &nb);
    if (reply.error != 0) {
        ret = -reply.error;
        goto fail;
    }
    if (reply.simple) {
        ret = -EINVAL;
        goto fail;
    }
    if (reply.type != NBD_REPLY_TYPE_BLOCK_STATUS) {
        ret = -EINVAL;
        goto fail;
    }

    if (client->meta_data_context_id != reply.context_id) {
        ret = -EINVAL;
        goto fail;
    }

    if (!(reply.flags & NBD_REPLY_FLAG_DONE)) {
        nbd_co_receive_reply(client, &request, &reply, NULL, NULL, NULL);
        if (reply.simple) {
            ret = -EINVAL;
            goto fail;
        }
        if (reply.error != 0) {
            ret = -reply.error;
            goto fail;
        }
        if (reply.type != NBD_REPLY_TYPE_NONE ||
            !(reply.flags & NBD_REPLY_FLAG_DONE)) {
            ret = -EINVAL;
            goto fail;
        }
    }

    *pextents = extents;
    *nb_extents = nb;
    nbd_coroutine_end(bs, &request);
    return 0;

fail:
    g_free(extents);
    nbd_coroutine_end(bs, &request);
    return ret;
}

/* nbd_client_co_load_bitmap_part() returns end of set area, i.e. first next
 * byte of unknown status (may be >= disk size, which means that the bitmap was
 * set up to the end).
 */
int64_t nbd_client_co_load_bitmap_part(BlockDriverState *bs, uint64_t offset,
                                       uint64_t bytes, BdrvDirtyBitmap *bitmap)
{
    int64_t ret;
    uint64_t start_byte;
    uint32_t nb_extents;
    int64_t i, start_sector, last_sector, nr_sectors;
    NBDExtent *extents = NULL;

    ret = nbd_client_co_cmd_block_status(bs, offset, bytes, &extents,
                                         &nb_extents);
    if (ret < 0) {
        return ret;
    }

    start_byte = offset;
    for (i = 0; i < nb_extents; ++i) {
        if (extents[i].flags == 1) {
            start_sector = start_byte >> BDRV_SECTOR_BITS;
            last_sector =
                (start_byte + extents[i].length - 1) >> BDRV_SECTOR_BITS;
            nr_sectors = last_sector - start_sector + 1;

            bdrv_set_dirty_bitmap(bitmap, start_sector, nr_sectors);
        }

        start_byte += extents[i].length;
    }

    g_free(extents);

    return ROUND_UP((uint64_t)start_byte,
                    (uint64_t)bdrv_dirty_bitmap_granularity(bitmap));
}

int64_t coroutine_fn nbd_client_co_get_block_status(BlockDriverState *bs,
                                                    int64_t sector_num,
                                                    int nb_sectors, int *pnum,
                                                    BlockDriverState **file)
{
    int64_t ret;
    uint32_t nb_extents;
    NBDExtent *extents;
    NBDClientSession *client = nbd_get_client_session(bs);

    if (!client->block_status_ok) {
        *pnum = nb_sectors;
        ret = BDRV_BLOCK_DATA | BDRV_BLOCK_ALLOCATED;
        if (bs->drv->protocol_name) {
            ret |= BDRV_BLOCK_OFFSET_VALID | (sector_num * BDRV_SECTOR_SIZE);
        }
        return ret;
    }

    ret = nbd_client_co_cmd_block_status(bs, sector_num << BDRV_SECTOR_BITS,
                                         nb_sectors << BDRV_SECTOR_BITS,
                                         &extents, &nb_extents);
    if (ret < 0) {
        return ret;
    }

    *pnum = extents[0].length >> BDRV_SECTOR_BITS;
    ret = (extents[0].flags & NBD_STATE_HOLE ? 0 : BDRV_BLOCK_ALLOCATED) |
          (extents[0].flags & NBD_STATE_ZERO ? BDRV_BLOCK_ZERO : 0);

    if ((ret & BDRV_BLOCK_ALLOCATED) && !(ret & BDRV_BLOCK_ZERO)) {
        ret |= BDRV_BLOCK_DATA;
    }

    g_free(extents);

    return ret;
}

void nbd_client_detach_aio_context(BlockDriverState *bs)
{
    NBDClientSession *client = nbd_get_client_session(bs);
    qio_channel_detach_aio_context(QIO_CHANNEL(client->ioc));
}

void nbd_client_attach_aio_context(BlockDriverState *bs,
                                   AioContext *new_context)
{
    NBDClientSession *client = nbd_get_client_session(bs);
    qio_channel_attach_aio_context(QIO_CHANNEL(client->ioc), new_context);
    aio_co_schedule(new_context, client->read_reply_co);
}

void nbd_client_close(BlockDriverState *bs)
{
    NBDClientSession *client = nbd_get_client_session(bs);
    NBDRequest request = { .type = NBD_CMD_DISC };

    if (client->ioc == NULL) {
        return;
    }

    nbd_send_request(client->ioc, &request);

    nbd_teardown_connection(bs);
}

int nbd_client_init(BlockDriverState *bs,
                    QIOChannelSocket *sioc,
                    const char *export,
                    QCryptoTLSCreds *tlscreds,
                    const char *hostname,
                    const char *bitmap_name,
                    Error **errp)
{
    NBDClientSession *client = nbd_get_client_session(bs);
    int ret;

    /* NBD handshake */
    logout("session init %s\n", export);
    qio_channel_set_blocking(QIO_CHANNEL(sioc), true, NULL);

    ret = nbd_receive_negotiate(QIO_CHANNEL(sioc), export,
                                &client->nbdflags,
                                tlscreds, hostname,
                                &client->ioc,
                                &client->size,
                                &client->structured_reply,
                                bitmap_name,
                                &client->bitmap_ok,
                                &client->block_status_ok, errp);
    if (ret < 0) {
        logout("Failed to negotiate with the NBD server\n");
        return ret;
    }
    if (client->nbdflags & NBD_FLAG_SEND_FUA) {
        bs->supported_write_flags = BDRV_REQ_FUA;
        bs->supported_zero_flags |= BDRV_REQ_FUA;
    }
    if (client->nbdflags & NBD_FLAG_SEND_WRITE_ZEROES) {
        bs->supported_zero_flags |= BDRV_REQ_MAY_UNMAP;
    }

    qemu_co_mutex_init(&client->send_mutex);
    qemu_co_queue_init(&client->free_sema);
    client->sioc = sioc;
    object_ref(OBJECT(client->sioc));

    if (!client->ioc) {
        client->ioc = QIO_CHANNEL(sioc);
        object_ref(OBJECT(client->ioc));
    }

    /* Now that we're connected, set the socket to be non-blocking and
     * kick the reply mechanism.  */
    qio_channel_set_blocking(QIO_CHANNEL(sioc), false, NULL);
    client->read_reply_co = qemu_coroutine_create(nbd_read_reply_entry, client);
    nbd_client_attach_aio_context(bs, bdrv_get_aio_context(bs));

    logout("Established connection with NBD server\n");
    return 0;
}
